/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"

// fonction de fusion, partagée par les différentes variantes de tri fusion
// fusionne a1[l..m[ et a1[m..r[ dans a2[l..r[
void merge(int a1[], int a2[], int l, int m, int r) {
  int i = l, j = m;
  for (int k = l; k < r; k++)
    if (i < m && (j == r || a1[i] <= a1[j]))
      a2[k] = a1[i++];
    else
      a2[k] = a1[j++];
}


// tri fusion récursif

void mergesortrec(int a[], int tmp[], int l, int r) {
  if (r - l <= 1) return;
  int m = l + (r - l) / 2;
  mergesortrec(a, tmp, l, m);
  mergesortrec(a, tmp, m, r);
  if (a[m-1] <= a[m]) return; // optim
  for (int i = l; i < r; i++) tmp[i] = a[i];
  merge(tmp, a, l, m, r);
}

void mergesort_(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  mergesortrec(a, tmp, 0, n);
  free(tmp);
}


// tri fusion ascendant (itératif)

int min(int x, int y) { return x < y ? x : y; }

void bottom_up_mergesort(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  for (int len = 1; len < n; len *= 2) {
    for (int i = 0; i < n; i++) tmp[i] = a[i];
    for (int k = 0; k < n - len; k += 2 * len) {
      merge(tmp, a, k, k + len, min(k + 2 * len, n));
    }
  }
  free(tmp);
}


// tri fusion utilisant les séquences croissantes déjà présentes

int find_run(int a[], int lo, int n) {
  while (++lo < n && a[lo-1] <= a[lo]);
  return lo;
}

void natural_mergesort(int a[], int n) {
  if (n <= 1) return;
  int *tmp = calloc(n, sizeof(int));
  while (true) {
    for (int lo = 0; lo < n-1; ) {
      // le premier run a[lo..mid[
      int mid = find_run(a, lo, n);
      if (mid == n) { if (lo == 0) return; break; }
      // le second run a[mid..hi[
      int hi = find_run(a, mid, n);
      // et on les fusionne
      for (int i = lo; i < hi; i++) tmp[i] = a[i];
      merge(tmp, a, lo, mid, hi);
      lo = hi;
    }
  }
}


// comptage des inversions d'un tableau

int merge_count(int a1[], int a2[], int l, int m, int r) {
  int i = l, j = m;
  int c = 0;
  for (int k = l; k < r; k++)
    if (i < m && (j == r || a1[i] <= a1[j]))
      a2[k] = a1[i++];
    else {
      a2[k] = a1[j++];
      c += m - i;
    }
  return c;
}

int mergesortrec_count(int a[], int tmp[], int l, int r) {
  if (r - l <= 1) return 0;
  int m = l + (r - l) / 2;
  int cleft = mergesortrec_count(a, tmp, l, m);
  int cright = mergesortrec_count(a, tmp, m, r);
  if (a[m-1] <= a[m]) return cleft + cright; // optim
  for (int i = l; i < r; i++) tmp[i] = a[i];
  int cmerge = merge_count(tmp, a, l, m, r);
  return cleft + cright + cmerge;
}

int mergesort_count(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  int c = mergesortrec_count(a, tmp, 0, n);
  free(tmp);
  return c;
}


